<?php

/**
 * Convert MOD files to NinjaTracker format.
 * 
 * by Jesse Blue / Ninjaforce
 *
 * 
 * July 28, 2018 - Initial release
 **/

/* the converter class, can be used on its own */

class ntconverter
{
    private $path_source_mod;        /* path to the MOD file that should be converted */
    private $path_output_ssm;        /* path to the resulting music (.ssm) file */
    private $path_output_w;          /* path to the resulting wave (.w) file */

    private $mod_data;               /* the binary file data of the MOD */
    private $max_number_instruments; /* how many instruments this MOD has at max */
    private $no_of_tracks;           /* how many tracks this MOD uses */
    private $max_pat_cnt;            /* number of patterns (blocks) that the MOD has */
    private $ptr_instnumb;           /* pointer to where instruments are in a MOD */
    private $mod_headlen;            /* length of the MOD header */
    private $ptr_songp_start;        /* pointer to where the Songinfo is, which block should be played after another */
    private $ptr_songp_end;          /* pointer to where the Songinfo ends */
    private $length_of_one_pattern;  /* how many bytes form a pattern in this MOD */
    private $length_of_all_patterns; /* how many bytes form all patterns in this MOD, combined */

    private $doc_wave;               /* the binary data that is the output wave file (always 64k) */
    private $instrument_information; /* array of instrument information from the MOD */
    private $instrument_data;        /* array of MOD instrument data */
    private $instrument_data_gs;     /* array of converted sample data */
    private $instrument_map;         /* array of which MOD instrument maps to which NT instrument */
    private $instrument_types;       /* array with type of gs instrument: 0 = not looped, 1 = looped, 2 = the loop itself */
    private $instrument_volume_gs;   /* array with volume of gs instrument */
    private $instrument_finetune_gs; /* array with finetune values of gs instrument */
    private $instrument_is_splitted; /* array of booleans about which MOD instrument is split into two */

    private $doc_ptrs;               /* array of pointers to where the GS instrument are placed in the DOC */
    private $doc_sizes;              /* array of size values that can be used by the DOC to play the instrument */

    private $ssm_notes;              /* array of note data for output ssm */
    private $ssm_sampleeffects;      /* array of sample and effect data for output ssm */
    private $ssm_effectvalues;       /* array of effect value data for output ssm */

    private $instrument_errors;      /* array of errors when converting instruments */
    private $pattern_errors;         /* array of errors when converting patterns */

    private static $ptr_first_instrument = 20;       /* pointer to where the MOD has its first instrument */
    private static $length_instrument_info = 30;     /* length of an instrument information block in the MOD */
    private static $offset_instrument_length = 22;   /* offset in instrument information block to the length of the instrument */
    private static $offset_instrument_finetune = 24; /* offset in instrument information block to the finetune value */
    private static $offset_instrument_volume = 25;   /* offset in instrument information block to the volume */
    private static $offset_instrument_rep_ptr = 26;  /* offset in instrument information block to the repeat pointer within the instrument */
    private static $offset_instrument_rep_len = 28;  /* offset in instrument information block to the length of the repeat */
    private static $ptr_mod_type = 1080;             /* pointer to where the MOD identifier code is */

    private static $ptr_by_type = [
        'old' => [ /* old trackers have up to 15 instruments */
            'max_number_instruments' => 15,
            'ptr_instnumb' => 950-480,
            'mod_headlen' => 1080-480,
            'ptr_songp_start' => 952-480,
            'ptr_songp_end' => 1080-480,
        ],
        'normal' => [ /* normal trackers have up to 31 instruments */
            'max_number_instruments' => 31,
            'ptr_instnumb' => 950,
            'mod_headlen' => 1084,
            'ptr_songp_start' => 952,
            'ptr_songp_end' => 1080,
        ],
    ];
    private static $length_of_note = 4;              /* 4 bytes represent a note in a mod */
    private static $number_of_lines_in_pattern = 64; /* there are always 64 lines in a pattern */
    private static $ssm_no_of_tracks = 14;           /* there are always 14 tracks in an ssm */
    private static $ssm_ptr_song = 470;              /* in an ssm, the list of blocks to play is located here */
    private static $max_ssm_instruments = 15;        /* there may be only a limited number of instruments in an ssm */
    private static $number_of_stopper_bytes = 8;     /* at the end of every instrument in DOC ram, add some 0 bytes to make the ensoniq stop playing */

    private static $mod_types = [ /* mapping for ID in MOD => number of tracks */
        'M.K.' => 4,
        '2CHN' => 2,
        '5CHN' => 5,
        '6CHN' => 6,
        '7CHN' => 7,
        '8CHN' => 8,
        '9CHN' => 9,
        '01CH' => 1,
        '02CH' => 2,
        '03CH' => 3,
        '04CH' => 4,
        '05CH' => 5,
        '06CH' => 6,
        '07CH' => 7,
        '08CH' => 8,
        '09CH' => 9,
        '10CH' => 10,
        '11CH' => 11,
        '12CH' => 12,
        '13CH' => 13,
        '14CH' => 14,
        '01CN' => 1,
        '02CN' => 2,
        '03CN' => 3,
        '04CN' => 4,
        '05CN' => 5,
        '06CN' => 6,
        '07CN' => 7,
        '08CN' => 8,
        '09CN' => 9,
        '10CN' => 10,
        '11CN' => 11,
        '12CN' => 12,
        '13CN' => 13,
        '14CN' => 14,
    ];

    private static $doc_page_boundaries = [ /* instruments in the DOC need to be aligned to this number of pages */
        1,
        2,
        4,
        8,
        16,
        32,
        64,
        128,
    ];

    private static $doc_register_sizes = [ /* the doc wavetable size register is set according to the wave size */
        0b0,
        0b001001,
        0b010010,
        0b011011,
        0b100100,
        0b101101,
        0b110110,
        0b111111,
    ];

    private static $mod_notes = [ /* music notes that are possible in the MOD */
        0,214,202,190,180,170,160,151,143,135,127,120,113, /* c5-b5 */
        428,404,381,360,339,320,302,285,269,254,240,226,   /* c4-b4 */
        856,808,762,720,678,640,604,570,538,508,480,453,   /* c3-b3 */
    ];

    private static $doc_notes = [ /* for each mod_note here is the corresponding DOC note */
        0,0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,
        0x24,0x25,0x26,0x27,0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
        0x18,0x19,0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,0x20,0x21,0x22,0x23,
    ];

    private static $mod_effects = [ /* effects that are possible in the MOD. This is not used, only for reference */
        0x00,0x0c,0x0a,0x0a,0x0f,
        0x01,0x02,0x03,0x04,0x07,0x08,0x09,0x0b,0x0d,0x0e,0x05,0x06
    ];

    private static $doc_effects = [ /* this is not used, only for reference */
        0x00,0x03,0x05,0x06,0x0f, /* supported by Soundsmith */
        0x01,0x02,0x0c,0x04,0x07,0x08,0x09,0x0b,0x0d,0x0e,0x0a,0x08 /* also supported by Ninjatracker */
    ];

    private static $doc_stereo_defaults = [
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
    ];

    /* ---- helper methods ---- */

    private function get_mod_file_byte(int $position)
    {
        return ord(substr($this->mod_data, $position, 1));
    }

    private function get_amiga_word(int $position)
    {
        return 256 * $this->get_mod_file_byte($position) + $this->get_mod_file_byte($position + 1);
    }

    private function word_to_gs_str($word)
    {
        $hi_byte = ($word & 0xff00) >> 8;
        $lo_byte = $word & 0x00ff;

        return chr($lo_byte) . chr($hi_byte);
    }

    private function get_smaller_length($length)
    {
        foreach(array_reverse(self::$doc_page_boundaries) as $page_boundary)
        {
            $byte_length = $page_boundary * 256 - self::$number_of_stopper_bytes;
            if($byte_length < $length)
            {
                return $byte_length;
            }
        }

        return $length;
    }

    /* ---- main program code ---- */

    public function __construct($path_source_mod, $path_output_ssm, $path_output_w)
    {
        $this->path_source_mod = $path_source_mod;
        $this->path_output_ssm = $path_output_ssm;
        $this->path_output_w = $path_output_w;

        $this->instrument_errors = [];
        $this->pattern_errors = [];
    }

    private function detect_mod_and_number_of_tracks()
    {
        /* find out if old or normal and how many tracks there are */
        $str_type = substr($this->mod_data, self::$ptr_mod_type, 4);
        if(isset(self::$mod_types[$str_type]))
        {
            $old_or_normal = 'normal';
            $this->no_of_tracks = self::$mod_types[$str_type];
        }
        else
        {
            $old_or_normal = 'old';
            $this->no_of_tracks = 4;
        }

        /* set values, depending on old or normal */
        $this->max_number_instruments = self::$ptr_by_type[$old_or_normal]['max_number_instruments'];
        $this->ptr_instnumb = self::$ptr_by_type[$old_or_normal]['ptr_instnumb'];
        $this->mod_headlen = self::$ptr_by_type[$old_or_normal]['mod_headlen'];
        $this->ptr_songp_start = self::$ptr_by_type[$old_or_normal]['ptr_songp_start'];
        $this->ptr_songp_end = self::$ptr_by_type[$old_or_normal]['ptr_songp_end'];
    }

    private function detect_number_of_patterns()
    {
        $max_pat_cnt = 0;
        for($pos = $this->ptr_songp_start; $pos < $this->ptr_songp_end; $pos++)
        {
            $byte = $this->get_mod_file_byte($pos);
            $max_pat_cnt = max($byte, $max_pat_cnt);
        }
        $max_pat_cnt++; /* add 1 because pattern 0 is also a pattern */

        $this->max_pat_cnt = $max_pat_cnt;
    }

    private function calc_pattern_length()
    {
        $this->length_of_one_pattern = $this->no_of_tracks * self::$length_of_note * self::$number_of_lines_in_pattern;

        $this->length_of_all_patterns = $this->length_of_one_pattern * $this->max_pat_cnt;
    }

    private function extract_instrument_information($instrument_number)
    {
        /* instrument numbers run from 1 to 15 or 1 to 31 */
        $ptr = self::$ptr_first_instrument + ($instrument_number - 1) * self::$length_instrument_info;

        $instrument_name = '';
        for($i = 0; $i < self::$length_instrument_info; $i++)
        {
            $byte = $this->get_mod_file_byte($ptr + $i);
            if($byte === 0)
            {
                break;
            }
            
            $instrument_name .= chr($byte);
        }

        $instrument_length = $this->get_amiga_word($ptr + self::$offset_instrument_length) * 2; /* given in words, transform to bytes */
        $instrument_volume = $this->get_mod_file_byte($ptr + self::$offset_instrument_volume);
        $instrument_finetune = (15 & $this->get_mod_file_byte($ptr + self::$offset_instrument_finetune)); /* lower nibble only */
        $instrument_rep_ptr = $this->get_amiga_word($ptr + self::$offset_instrument_rep_ptr) * 2; /* given in words, transform to bytes */
        $instrument_rep_len = $this->get_amiga_word($ptr + self::$offset_instrument_rep_len) * 2; /* given in words, transform to bytes */

        return [
            'name' => $instrument_name,
            'length' => $instrument_length,
            'finetune' => $instrument_finetune,
            'volume' => $instrument_volume,
            'rep_ptr' => $instrument_rep_ptr,
            'rep_len' => $instrument_rep_len,
        ];
    }

    private function extract_all_instrument_information()
    {
        $this->instrument_information = [];

        for($i = 1; $i <= $this->max_number_instruments; $i++)
        {
            $this->instrument_information[$i] = $this->extract_instrument_information($i);
        }
    }

    private function extract_instrument_data($instrument_number)
    {
        /* calculcate the pointer in the MOD where the instrument begins */
        $ptr = $this->mod_headlen + $this->length_of_all_patterns;

        for($i = 1; $i < $instrument_number; $i++)
        {
            $ptr += $this->instrument_information[$i]['length'];
        }

        /* get instrument data */
        $instrument_data = [];
        for($i = $ptr; $i < $ptr + $this->instrument_information[$instrument_number]['length']; $i++)
        {
            $instrument_data[] = $this->get_mod_file_byte($i);
        }
        return $instrument_data;
    }

    private function extract_all_instrument_data()
    {
        $this->instrument_data = [];

        for($i = 1; $i <= $this->max_number_instruments; $i++)
        {
            if($this->instrument_information[$i]['length'] > 0)
            {
                $this->instrument_data[$i] = $this->extract_instrument_data($i);
            }
        }
    }

    private function convert_instrument_data($mod_instrument_number, $offset, $length, $loop = false)
    {
        $instrument_data_gs = [];

        for($i = $offset; $i < $offset + $length; $i++)
        {
            $byte = $this->instrument_data[$mod_instrument_number][$i];

            /* shift waveform, the GS has 128 for silence, the Amiga has 0 */
            $ensoniq_byte = 255 & ($byte + 128);

            /* make sure we do not have a 0 in the wave, which would stop the ensoniq from playing */
            if($ensoniq_byte === 0)
            {
                $ensoniq_byte = 1;
            }

            $instrument_data_gs[] = $ensoniq_byte;
        }

        /* when we're dealing with the loop of an instrument and the length is < 256, copy the loop to improve the sound quality */
        /*if($loop)
        {
            $loop = $instrument_data_gs;
            while(count($loop) < 512 - self::$number_of_stopper_bytes)
            {
                $loop = array_merge($loop, $instrument_data_gs);
            }
            $instrument_data_gs = $loop;
        }*/

        /* add stopper bytes */
        for($i = 0; $i < self::$number_of_stopper_bytes; $i++)
        {
            $instrument_data_gs[] = 0;
        }

        return $instrument_data_gs;
    }

    private function add_gs_instrument($mod_instrument_number, $instrument_data_gs, $instrument_type)
    {
        $index = 1 + count($this->instrument_data_gs);

        $this->instrument_data_gs[$index] = $instrument_data_gs;
        
        /* Add mapping.
           Because we maybe split looped instruments into two for the GS version,
           we need to map which MOD instrument is which NT instrument.
           But for the mapping, we only need the first instrument.
           Because if the instrument consists of a head and a loop, the head
           is always first and loop is second.
        */
        if(!isset($this->instrument_map[$mod_instrument_number]))
        {
            $this->instrument_map[$mod_instrument_number] = $index;
        }

        $this->instrument_types[$index] = $instrument_type;
        $this->instrument_volume_gs[$index] = min(255, $this->instrument_information[$mod_instrument_number]['volume'] * 4);
        $this->instrument_finetune_gs[$index] = $this->instrument_information[$mod_instrument_number]['finetune'];
    }

    private function convert_instruments()
    {
        $this->instrument_data_gs = [];
        $this->instrument_map = [];

        for($i = 1; $i <= $this->max_number_instruments; $i++)
        {
            $length = $this->instrument_information[$i]['length'];

            if($length === 0)
            {
                /* No instrument, skip */
                continue;
            }

            $rep_ptr = $this->instrument_information[$i]['rep_ptr'];
            $rep_len = $this->instrument_information[$i]['rep_len'];

            $instrument_is_splitted = false;

            if($rep_len < 3)
            {
                /* instrument is not looped */
                $instrument_data_gs_sample = $this->convert_instrument_data($i, 0, $length);
                $this->add_gs_instrument($i, $instrument_data_gs_sample, 0);
            }
            else
            {
                /* instrument is looped */
                /* if the instrument consists of a head and a loop, save each as individual instrument */
                /* but if the instrument header is very short (< 256 bytes), save only the loop */
                if($rep_ptr >= 256)
                {
                    $instrument_data_gs_head = $this->convert_instrument_data($i, 0, $rep_ptr, false);
                    $this->add_gs_instrument($i, $instrument_data_gs_head, 1);

                    $instrument_data_gs_loop = $this->convert_instrument_data($i, $rep_ptr, $rep_len, true);
                    $this->add_gs_instrument($i, $instrument_data_gs_loop, 2);

                    $instrument_is_splitted = true;
                }
                else
                {
                    $instrument_data_gs_loop = $this->convert_instrument_data($i, $rep_ptr, $rep_len, true);
                    $this->add_gs_instrument($i, $instrument_data_gs_loop, 3);
                }
            }

            $this->instrument_is_splitted[$i] = $instrument_is_splitted;
        }
    }

    private function check_instruments()
    {
        if(count($this->instrument_data_gs) > 15)
        {
            $this->instrument_errors[] = 'Too many instruments:';
            $this->instrument_errors[] = '- Ninjatracker only allows up to 15 instruments.';

            $number_of_instruments = count($this->instrument_data);
            if($number_of_instruments > 15)
            {
                $this->instrument_errors[] = sprintf('- The MOD already has %s instruments.', $number_of_instruments);
            }
            else
            {
                $this->instrument_errors[] = sprintf('- The MOD has %s instruments.', $number_of_instruments);
            }

            $number_splitted = 0;
            foreach($this->instrument_is_splitted as $index => $is_splitted)
            {
                $number_splitted += $is_splitted ? 1 : 0;
            }

            if($number_splitted > 0)
            {
                $this->instrument_errors[] = sprintf('- Because of looped instruments, the converter creates %s additional instrument%s.', 
                    $number_splitted,
                    ($number_splitted != 1) ? 's' : ''
                );
            }

            $this->instrument_errors[] = '- Consider getting rid of instruments or try removing instrument loop headers.';
        }
    }

    private function init_docwave()
    {
        $this->doc_wave = array_fill(0, 65536, 0); /* init with 0 */
    }
    
    private function sort_instruments_by_size()
    {
        /* sort instruments by size, descending */
        $sizes = [];
        foreach($this->instrument_data_gs as $instrument_number_gs => $instrument_data_gs)
        {
            $length = count($instrument_data_gs);
            $page_size = ceil($length / 256); /* how many pages does our instrument need? */
            $align_boundary = 0;
            foreach(self::$doc_page_boundaries as $page_limit)
            {
                if($page_size <= $page_limit)
                {
                    $align_boundary = $page_limit;
                    break;
                }
            }
            $sizes[$instrument_number_gs] = [$length, $page_size, $align_boundary];
        }

        $all_sizes = $sizes;
        $sorted_sizes = [];
        foreach($all_sizes as $dummy)
        {
            $arr_largest_size = [0, 0, 0];
            $largest_instrument_number = 0;

            foreach($sizes as $instrument_number_gs => $arr_size)
            {
                list($length, $page_size, $my_align_boundary) = $arr_size;
                if($my_align_boundary >= $arr_largest_size[2])
                {
                    $largest_instrument_number = $instrument_number_gs;
                    $arr_largest_size = $arr_size;
                }
            }

            $sizes[$largest_instrument_number] = 0;
            $sorted_sizes[$largest_instrument_number] = $arr_largest_size;
        }

        return $sorted_sizes;
    }

    private function get_instrument_error($instrument_number_gs, $error)
    {
        /* from the GS instrument number, find the MOD instrument number */
        $instrument_map_reverse = array_flip($this->instrument_map);

        if(isset($instrument_map_reverse[$instrument_number_gs]))
        {
            $instrument_number_mod = $instrument_map_reverse[$instrument_number_gs];
        }
        else
        {
            $instrument_number_mod = $instrument_map_reverse[$instrument_number_gs - 1];
        }

        /* Give a hint about making the instrument smaller */
        $sizehint = '';
        switch($this->instrument_types[$instrument_number_gs])
        {
            default:
            case 0:
                $length = $this->instrument_information[$instrument_number_mod]['length'];
                $sizehint = 'consider reducing the size from %s to %s bytes';
                break;
            case 1:
                $length = count($this->instrument_data_gs[$instrument_number_gs]);
                $sizehint = 'consider reducing the loop header from %s to %s bytes or removing it';
                break;
            case 2:
                $length = count($this->instrument_data_gs[$instrument_number_gs]);
                $sizehint = 'consider reducing the loop from %s to %s bytes';
                break;
        }

        $smaller_length = $this->get_smaller_length($length);
        if($smaller_length < $length)
        {
            $sizehint = ' - ' . sprintf($sizehint, $length, $smaller_length);
        }

        return sprintf('Instrument %s: %s%s.', str_pad($instrument_number_mod, 2, '0', STR_PAD_LEFT), $error, $sizehint);
    }

    private function put_instruments_in_docwave()
    {   
        /* for each page (a 256 byte block), we mark if it is used or not; init with 0 */
        $doc_pages = array_fill(0, 256, 0);
        
        /* loop over instruments */
        $sorted_sizes = $this->sort_instruments_by_size();
        foreach($sorted_sizes as $instrument_number_gs => $arr_size)
        {
            list($length, $page_size, $my_align_boundary) = $arr_size;

            /* Depending on the size of an instrument, find the matching alignment.
               For example, an instrument with length of 423 takes up 2 pages (=512 bytes) and therefore needs to be aligned on 512 bytes.
             */
            $align_boundary = 0;
            foreach(self::$doc_page_boundaries as $i => $page_limit)
            {
                if($page_size <= $page_limit)
                {
                    $align_boundary = $page_limit;
                    $doc_register_size = self::$doc_register_sizes[$i];
                    break;
                }
            }

            /* If we don't find a matching alignment, it means that the instrument is larger than 32k and the DOC can't play that */
            if($align_boundary === 0)
            {
                $error = 'Too large (must be < 32k)';
                $this->instrument_errors[] = $this->get_instrument_error($instrument_number_gs, $error);
                continue;
            }

            /* find a free slot in the DOC ram */
            $step_size = $align_boundary;
            $page_start = -1;
            for($i = 0; $i < 256; $i += $step_size)
            {
                if($doc_pages[$i] === 0)
                {
                    /* we found a free slot */
                    $page_start = $i;
                    break;
                }
            }

            if($page_start === -1)
            {
                $error = 'Too large';
                $this->instrument_errors[] = $this->get_instrument_error($instrument_number_gs, $error);
                continue;
            }

            /* copy instrument data into DOC */
            $ptr_in_doc = $page_start * 256;
            foreach($this->instrument_data_gs[$instrument_number_gs] as $byte)
            {
                $this->doc_wave[$ptr_in_doc] = $byte;
                $ptr_in_doc++;
            }

            /* record where we place the instrument */
            $this->doc_ptrs[$instrument_number_gs] = $page_start;

            /* record the size of the instrument as the doc needs it */
            $this->doc_sizes[$instrument_number_gs] = $doc_register_size;

            /* mark DOC ram as occupied */
            for($i = $page_start; $i < $page_start + $page_size; $i++)
            {
                $doc_pages[$i] = 1;
            }
        }
    }

    private function save_docwave()
    {
        $str = '';

        /* first 2 bytes are the instrument count */
        $instrument_count_gs = count($this->instrument_data_gs);
        $str .= chr($instrument_count_gs);
        $str .= chr(0);

        /* add the wave data */
		foreach($this->doc_wave as $b)
		{
			$str .= chr($b);
        }
        
        /* then there are 32 zero bytes */
        $str .= str_repeat(chr(0), 0x20);

        /* then write doc registers */
        for($instrument_number_gs = 1; $instrument_number_gs <= $instrument_count_gs; $instrument_number_gs++)
        {
            //TopKey             Byte      The highest MIDI semitone this waveform will play.
            //WaveAddress        Byte      The high byte of the waveform address.
            //WaveSize           Byte      Sets both the size of the wavetable and the
            //DOCMode            Byte      Placed in the DOCs Mode register.
            //RelPitch           Word      Used to tune the waveform.
            $str .= chr(127);
            $str .= chr($this->doc_ptrs[$instrument_number_gs]);
            $str .= chr($this->doc_sizes[$instrument_number_gs]);
            $str .= chr(2);
            $str .= chr(0);
            $str .= chr(0);
            
            $str .= chr(127);
            $str .= chr($this->doc_ptrs[$instrument_number_gs]);
            $str .= chr($this->doc_sizes[$instrument_number_gs]);
            $str .= chr(2);
            $str .= chr(0);
            $str .= chr(0);

            //fill with zero bytes
            $str .= str_repeat(chr(0), 80);
        }

        $str .= str_repeat(chr(0), 0x3c + 0x20 - 2);
		
		file_put_contents($this->path_output_w, $str);
    }

    private function convert_period($period)
    {
        foreach(self::$mod_notes as $index => $mod_note)
        {
            if($period == $mod_note)
            {
                return self::$doc_notes[$index];
            }
        }

        throw new Exception(sprintf('Unable to convert note %s. Only classic ProTracker octaves are supported.', $period));
    }

    private function convert_effect_number($effect_number)
    {
        return $effect_number;

        /* for soundsmith, we need to do some modifications here, but ninjatracker uses protracker effects
        foreach(self::$mod_effects as $index => $mod_effect)
        {
            if($effect_number == $mod_effect)
            {
                return self::$doc_effects[$index];
            }
        }

        throw new Exception(sprintf('Unable to convert effect %s', $effect_number));
        */
    }

    private function convert_instrument_number($mod_instrument_number)
    {
        if($mod_instrument_number === 0)
        {
            return 0; /* no instrument */
        }

        if(isset($this->instrument_map[$mod_instrument_number]))
        {
            return $this->instrument_map[$mod_instrument_number];
        }

        throw new Exception(sprintf('Instrument %s not found!', $mod_instrument_number));
    }

    private function convert_mod_note($ptr_to_note)
    {
        /* 4 bytes are one note/effect */
        $word0 = $this->get_amiga_word($ptr_to_note);
        $word1 = $this->get_amiga_word($ptr_to_note + 2);

        /* extract mod data */
        $mod_period = $word0 & 0x0fff;
        $mod_instrument_number = (($word0 & 0xf000) >> 8) + (($word1 & 0xf000) >> 12);
        $mod_effect_number = ($word1 & 0x0f00) >> 8;
        $mod_effect_value = ($word1 & 0x00ff);

        /* convert to gs */
        $gs_note = $this->convert_period($mod_period);
        $gs_instrument_number = $this->convert_instrument_number($mod_instrument_number);
        $gs_effect_number = $this->convert_effect_number($mod_effect_number);
        $gs_effect_value = $mod_effect_value; /* converting to SoundSmith would require some changes here */

        return [$gs_note, $gs_instrument_number, $gs_effect_number, $gs_effect_value];
    }

    private function add_note_to_ssm(array $note_data)
    {
        list($gs_note, $gs_instrument_number, $gs_effect_number, $gs_effect_value) = $note_data;

        /* sanity checks */
        if($gs_note > 255)
        {
            throw new Exception('Ninjatracker does not allow a note value > 255!');
        }
        if($gs_instrument_number > 15)
        {
            throw new Exception('Ninjatracker does not allow more than 15 instruments!');
        }
        if($gs_effect_number > 15)
        {
            throw new Exception('Ninjatracker does not allow effect numbers > 15!');
        }
        if($gs_effect_value > 255)
        {
            throw new Exception('Ninjatracker does not allow effect values > 255!');
        }

        $this->ssm_notes[] = $gs_note;
        $this->ssm_sampleeffects[] = $gs_instrument_number * 16 + $gs_effect_number;
        $this->ssm_effectvalues[] = $gs_effect_value;
    }

    private function convert_pattern($pattern_number)
    {
        $pattern_len = $this->no_of_tracks * self::$number_of_lines_in_pattern * self::$length_of_note;
        $pattern_offset = $this->mod_headlen + $pattern_len * $pattern_number;
        
        try {
            for($line = 0; $line < self::$number_of_lines_in_pattern; $line++)
            {
                $line_offset = $line * $this->no_of_tracks * self::$length_of_note;
    
                for($track = 0; $track < $this->no_of_tracks; $track++)
                {
                    $track_offset = $track * self::$length_of_note;
    
                    $ptr = $pattern_offset + $line_offset + $track_offset;
    
                    $ssm_note_data = $this->convert_mod_note($ptr);
                    $this->add_note_to_ssm($ssm_note_data);
                }

                while($track < self::$ssm_no_of_tracks)
                {
                    $this->add_note_to_ssm([0,0,0,0]); /* add empty notes to fill up all ssm tracks */
                    $track++;
                }
            }
        } catch(Exception $ex)
        {
            $this->pattern_errors[] = sprintf('Error in pattern %s, line %s, track %s: %s',
                str_pad($pattern_number, 2, '0', STR_PAD_LEFT),
                str_pad($line, 2, '0', STR_PAD_LEFT),
                str_pad($track + 1, 2, '0', STR_PAD_LEFT),
                $ex->getMessage()
            );
        }
    }

    private function convert_patterns()
    {
        for($pattern_number = 0; $pattern_number < $this->max_pat_cnt; $pattern_number++)
        {
            $this->convert_pattern($pattern_number);
        }
    }

    private function create_ssm_instrument_data($instrument_name, $instrument_volume, $instrument_type, $instrument_finetuning)
    {
        $max_instrument_name_length = 23;

        $instrument_name = substr($instrument_name, 0, $max_instrument_name_length); /* enforce character limit for instrument name */
        $instrument_name_padded = str_pad($instrument_name, $max_instrument_name_length, chr(0), STR_PAD_RIGHT); /* fill up with zeros */
        
        $str = '';
        $str .= chr(strlen($instrument_name)); /* 1 byte for length */
        $str .= $instrument_name_padded;       /* 23 bytes name filled up with zeros */
        $str .= chr($instrument_volume);       /* 1 byte volume */
        $str .= chr(0);                        /* 1 byte zero */
        $str .= chr($instrument_type);         /* 1 byte instrument type */
        $str .= chr($instrument_finetuning);   /* 1 byte instrument finetuning */
        $str .= chr(0);                        /* 1 byte zero */
        $str .= chr(0);                        /* 1 byte zero */

        return $str; /* always 30 bytes */
    }

    private function save_ssm()
    {
        /* identifier */
        $str = 'SONGOK';

        /* 2 bytes block length */
        $ssm_blocklen = count($this->ssm_notes);
        $str .= $this->word_to_gs_str($ssm_blocklen);

        /* 2 bytes tempo, always 6 */
        $str .= $this->word_to_gs_str(6);

        /* 10 padding bytes */
        $str .= str_repeat(chr(0), 10);

        /* 450 bytes instrument data, 30 bytes for every instrument */
        for($instrument_number_gs = 1; $instrument_number_gs <= self::$max_ssm_instruments; $instrument_number_gs++)
        {
            if(isset($this->instrument_data_gs[$instrument_number_gs]))
            {
                /* if available, create instrument data */
                $instrument_name = 'N'; /* this is a bug in the old modconvert */
                $instrument_volume = $this->instrument_volume_gs[$instrument_number_gs];
                $instrument_type = $this->instrument_types[$instrument_number_gs];
                $instrument_finetuning = $this->instrument_finetune_gs[$instrument_number_gs];

                $str .= $this->create_ssm_instrument_data($instrument_name, $instrument_volume, $instrument_type, $instrument_finetuning);
            }
            else
            {
                /* if not, fill with zeros */
                $str .= str_repeat(chr(0), 30);
            }
        }

        /* add song data */
        for($i = 0; $i < 130; $i++)
        {
            $ptr_mod = $this->ptr_instnumb + $i;
            $byte = $this->get_mod_file_byte($ptr_mod);

            $str .= chr($byte);
        }

        /* add note data */
        foreach($this->ssm_notes as $byte)
        {
            $str .= chr($byte);
        }

        /* add sample and effect data */
        foreach($this->ssm_sampleeffects as $byte)
        {
            $str .= chr($byte);
        }

        /* add effect value data */
        foreach($this->ssm_effectvalues as $byte)
        {
            $str .= chr($byte);
        }

        /* 30 bytes stereo data */
        for($i = 1; $i <= 15; $i++)
        {
            $channel_number = self::$doc_stereo_defaults[$i - 1];
            $str .= $this->word_to_gs_str($channel_number);
        }

        file_put_contents($this->path_output_ssm, $str);
    }

    private function load_mod()
    {
        $this->mod_data = file_get_contents($this->path_source_mod);

        if($this->mod_data === false)
        {
            throw new Exception(sprintf('Unable to load MOD %s', $this->path_source_mod), self::FATAL_ERROR);
        }
    }

    private function check_no_errors_occurred()
    {
        $no_errors = empty($this->instrument_errors) && empty($this->pattern_errors);
        if($no_errors)
        {
            return true;
        }

        $show_max_pattern_errors = 10;
        $number_of_pattern_errors = count($this->pattern_errors);

        $all_errors = $this->instrument_errors;

        $all_errors = array_merge($all_errors, array_slice($this->pattern_errors, 0, $show_max_pattern_errors));

        if($number_of_pattern_errors > $show_max_pattern_errors)
        {
            $all_errors[] = sprintf('... (%s more pattern errors)', $number_of_pattern_errors - $show_max_pattern_errors);
        }
        
        throw new Exception(implode($all_errors, PHP_EOL));
    }

    public function convert()
    {
        try {
            /* load mod */
            $this->load_mod();

            /* read basic information from mod */
            $this->detect_mod_and_number_of_tracks();
            $this->detect_number_of_patterns();
            $this->calc_pattern_length();
    
            /* convert instruments */
            $this->extract_all_instrument_information();
            $this->extract_all_instrument_data();
            $this->convert_instruments();
            $this->check_instruments();
            $this->init_docwave();
            $this->put_instruments_in_docwave();
    
            /* convert patterns */
            $this->convert_patterns();
            
            /* save, if there are no errors */
            if($this->check_no_errors_occurred())
            {
                $this->save_docwave();
                $this->save_ssm();
            }

            $result = ['success' => true];
        }
        catch(Exception $ex)
        {
            $result = ['success' => false, 'message' => $ex->getMessage(). PHP_EOL];
        }
        
        return $result;
    }
}


/* the command line class to run the converter */

class run_converter
{
    public function parse_commandline()
    {
        if(isset($_SERVER['argv'][1]))
        {
            $mod_arg = $_SERVER['argv'][1];
            if(!file_exists($mod_arg))
            {
                throw new Exception(sprintf('Unable to find MOD file %s', $mod_arg));
            }
        }
        else
        {
            throw new Exception('Please specify a MOD file!');
        }

        if(strtoupper(substr($mod_arg,-4)) === '.MOD')
        {
            $name = substr($mod_arg, 0, -4);
        }
        else
        {
            $name = $mod_arg;
        }

        $this->path_source_mod = $mod_arg;
        $this->path_output_ssm = $name . '.ssm';
        $this->path_output_w = $name . '.w';
    }

    public function run()
    {
        $return_code = 0;

        try {
            $this->parse_commandline();

            $ntconverter = new ntconverter($this->path_source_mod, $this->path_output_ssm, $this->path_output_w);
            $result = $ntconverter->convert();
    
            if($result['success'] === false)
            {
                throw new Exception($result['message']);
            }
        }
        catch(Exception $ex)
        {
            echo 'UNABLE TO CONVERT:' . PHP_EOL . $ex->getMessage(). PHP_EOL;
            $return_code = 255;
        }
        
        return $return_code;
    }
}

$ntconverter = new run_converter();
exit($ntconverter->run());
